@@ -49,6 +49,7 @@ struct _WnckWindowPrivate
   WnckScreen *screen;
   WnckApplication *app;
   Window group_leader;
+  Window transient_for;
   char *name;
   char *icon_name;
   char *session_id;
@@ -56,6 +57,8 @@ struct _WnckWindowPrivate
   int pid;
   int workspace;
 
+  WnckWindowType wintype;
+  
   GdkPixbuf *icon;
   GdkPixbuf *mini_icon;
 
@@ -77,6 +80,13 @@ struct _WnckWindowPrivate
   guint skip_taskbar : 1;
   guint is_sticky : 1;
 
+  /* _NET_WM_STATE_HIDDEN doesn't map directly into an
+   * externally-visible state (it determines the WM_STATE
+   * interpretation)
+   */
+  guint net_wm_state_hidden : 1;
+  guint wm_state_iconic : 1;
+  
   /* idle handler for updates */
   guint update_handler;
 
@@ -85,11 +95,13 @@ struct _WnckWindowPrivate
    */
   guint need_update_name : 1;
   guint need_update_state : 1;
-  guint need_update_minimized : 1;
+  guint need_update_wm_state : 1;
   guint need_update_icon_name : 1;
   guint need_update_workspace : 1;
   guint need_emit_icon_changed : 1;
   guint need_update_actions : 1;
+  guint need_update_wintype : 1;
+  guint need_update_transient_for : 1;
 };
 
 enum {
@@ -119,10 +131,12 @@ static void emit_geometry_changed  (WnckWindow      *window);
 
 static void update_name      (WnckWindow *window);
 static void update_state     (WnckWindow *window);
-static void update_minimized (WnckWindow *window);
+static void update_wm_state  (WnckWindow *window);
 static void update_icon_name (WnckWindow *window);
 static void update_workspace (WnckWindow *window);
 static void update_actions   (WnckWindow *window);
+static void update_wintype   (WnckWindow *window);
+static void update_transient_for (WnckWindow *window);
 static void unqueue_update   (WnckWindow *window);
 static void queue_update     (WnckWindow *window);
 static void force_update_now (WnckWindow *window);
@@ -339,7 +353,8 @@ _wnck_window_create (Window      xwindow,
   window->priv->pid =
     _wnck_get_pid (window->priv->xwindow);
 
-  _wnck_get_window_geometry (xwindow,
+  _wnck_get_window_geometry (WNCK_SCREEN_XSCREEN (window->priv->screen),
+			     xwindow,
                              &window->priv->x,
                              &window->priv->y,
                              &window->priv->width,
@@ -348,9 +363,11 @@ _wnck_window_create (Window      xwindow,
   window->priv->need_update_name = TRUE;
   window->priv->need_update_state = TRUE;
   window->priv->need_update_icon_name = TRUE;
-  window->priv->need_update_minimized = TRUE;
+  window->priv->need_update_wm_state = TRUE;
   window->priv->need_update_workspace = TRUE;
   window->priv->need_update_actions = TRUE;
+  window->priv->need_update_wintype = TRUE;
+  window->priv->need_update_transient_for = TRUE;
   force_update_now (window);
 
   return window;
@@ -498,6 +515,22 @@ wnck_window_get_pid (WnckWindow *window)
   return window->priv->pid;
 }
 
+/**
+ * wnck_window_get_window_type:
+ * @window: a #WnckWindow
+ * 
+ * Retrieves the semantic type of the window.
+ * 
+ * Return value: semantic type
+ **/
+WnckWindowType
+wnck_window_get_window_type (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), 0);
+  
+  return window->priv->wintype;
+}
+
 /**
  * wnck_window_is_minimized:
  * @window: a #WnckWindow
@@ -599,7 +632,8 @@ wnck_window_close (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_close (window->priv->xwindow);
+  _wnck_close (WNCK_SCREEN_XSCREEN (window->priv->screen),
+	       window->priv->xwindow);
 }
 
 void
@@ -623,7 +657,8 @@ wnck_window_maximize                (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"),
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"));
@@ -634,7 +669,8 @@ wnck_window_unmaximize              (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"),
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"));
@@ -645,7 +681,8 @@ wnck_window_maximize_horizontally   (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"),
                       0);
@@ -656,7 +693,8 @@ wnck_window_unmaximize_horizontally (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"),
                       0);
@@ -667,7 +705,8 @@ wnck_window_maximize_vertically     (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"),
                       0);
@@ -678,7 +717,8 @@ wnck_window_unmaximize_vertically   (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"),
                       0);
@@ -689,7 +729,8 @@ wnck_window_shade                   (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_SHADED"),
                       0);
@@ -700,7 +741,8 @@ wnck_window_unshade                 (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_SHADED"),
                       0);
@@ -711,7 +753,8 @@ wnck_window_stick                   (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_STICKY"),
                       0);
@@ -722,7 +765,8 @@ wnck_window_unstick                 (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_STICKY"),
                       0);
@@ -744,7 +788,7 @@ wnck_window_get_workspace (WnckWindow *window)
   if (window->priv->workspace == ALL_WORKSPACES)
     return NULL;
   else
-    return wnck_workspace_get (window->priv->workspace);
+    return wnck_screen_get_workspace (window->priv->screen, window->priv->workspace);
 }
 
 void
@@ -754,7 +798,8 @@ wnck_window_move_to_workspace (WnckWindow    *window,
   g_return_if_fail (WNCK_IS_WINDOW (window));
   g_return_if_fail (WNCK_IS_WORKSPACE (space));
 
-  _wnck_change_workspace (window->priv->xwindow,
+  _wnck_change_workspace (WNCK_SCREEN_XSCREEN (window->priv->screen),
+			  window->priv->xwindow,
                           wnck_workspace_get_number (space));
 }
 
@@ -781,7 +826,8 @@ wnck_window_pin (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_workspace (window->priv->xwindow,
+  _wnck_change_workspace (WNCK_SCREEN_XSCREEN (window->priv->screen),
+			  window->priv->xwindow,
                           ALL_WORKSPACES);
 }
 
@@ -808,7 +854,8 @@ wnck_window_unpin (WnckWindow *window)
 
   active = wnck_screen_get_active_workspace (window->priv->screen);
 
-  _wnck_change_workspace (window->priv->xwindow,
+  _wnck_change_workspace (WNCK_SCREEN_XSCREEN (window->priv->screen),
+			  window->priv->xwindow,
                           active ? wnck_workspace_get_number (active) : 0);
 }
 
@@ -824,7 +871,8 @@ wnck_window_activate (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_activate (window->priv->xwindow);
+  _wnck_activate (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		  window->priv->xwindow);
 }
 
 /**
@@ -861,11 +909,6 @@ get_icons (WnckWindow *window)
                         DEFAULT_MINI_ICON_HEIGHT))
     {
       window->priv->need_emit_icon_changed = TRUE;
-      
-      if (icon)
-        g_object_ref (G_OBJECT (icon));
-      if (mini_icon)
-        g_object_ref (G_OBJECT (mini_icon));
 
       if (window->priv->icon)
         g_object_unref (G_OBJECT (window->priv->icon));
@@ -1088,7 +1131,7 @@ _wnck_window_process_property_notify (WnckWindow *window,
   else if (xevent->xproperty.atom ==
       _wnck_atom_get ("WM_STATE"))
     {
-      window->priv->need_update_minimized = TRUE;
+      window->priv->need_update_wm_state = TRUE;
       queue_update (window);
     }
   else if (xevent->xproperty.atom ==
@@ -1123,12 +1166,25 @@ _wnck_window_process_property_notify (WnckWindow *window,
       window->priv->need_update_workspace = TRUE;
       queue_update (window);
     }
+  else if (xevent->xproperty.atom ==
+           _wnck_atom_get ("_NET_WM_WINDOW_TYPE"))
+    {
+      window->priv->need_update_wintype = TRUE;
+      queue_update (window);
+    }
+  else if (xevent->xproperty.atom ==
+           _wnck_atom_get ("WM_TRANSIENT_FOR"))
+    {
+      window->priv->need_update_transient_for = TRUE;
+      window->priv->need_update_wintype = TRUE;
+      queue_update (window);
+    }
   else if (xevent->xproperty.atom ==
            _wnck_atom_get ("_NET_WM_ICON") ||
            xevent->xproperty.atom ==
            _wnck_atom_get ("KWM_WIN_ICON") ||
            xevent->xproperty.atom ==
-           _wnck_atom_get ("WM_NORMAL_HINTS"))
+           _wnck_atom_get ("WM_HINTS"))
     {
       _wnck_icon_cache_property_changed (window->priv->icon_cache,
                                          xevent->xproperty.atom);
@@ -1147,7 +1203,8 @@ _wnck_window_process_configure_notify (WnckWindow *window,
     }
   else
     {
-      _wnck_get_window_position (window->priv->xwindow,
+      _wnck_get_window_position (WNCK_SCREEN_XSCREEN (window->priv->screen),
+				 window->priv->xwindow,
                                  &window->priv->x,
                                  &window->priv->y);
     }
@@ -1159,21 +1216,21 @@ _wnck_window_process_configure_notify (WnckWindow *window,
 }
 
 static void
-update_minimized (WnckWindow *window)
+update_wm_state (WnckWindow *window)
 {
   int state;
 
-  if (!window->priv->need_update_minimized)
+  if (!window->priv->need_update_wm_state)
     return;
 
-  window->priv->need_update_minimized = FALSE;
+  window->priv->need_update_wm_state = FALSE;
 
-  window->priv->is_minimized = FALSE;
+  window->priv->wm_state_iconic = FALSE;
 
   state = _wnck_get_wm_state (window->priv->xwindow);
 
   if (state == IconicState)
-    window->priv->is_minimized = TRUE;
+    window->priv->wm_state_iconic = TRUE;
 }
 
 static void
@@ -1182,45 +1239,99 @@ update_state (WnckWindow *window)
   Atom *atoms;
   int n_atoms;
   int i;
+  gboolean reread_net_wm_state;
 
-  if (!window->priv->need_update_state)
-    return;
+  reread_net_wm_state = window->priv->need_update_state;
 
   window->priv->need_update_state = FALSE;
 
-  window->priv->is_maximized_horz = FALSE;
-  window->priv->is_maximized_vert = FALSE;
-  window->priv->is_sticky = FALSE;
-  window->priv->is_shaded = FALSE;
-  window->priv->skip_taskbar = FALSE;
-  window->priv->skip_pager = FALSE;
-
-  atoms = NULL;
-  n_atoms = 0;
-  _wnck_get_atom_list (window->priv->xwindow,
-                       _wnck_atom_get ("_NET_WM_STATE"),
-                       &atoms, &n_atoms);
-
-  i = 0;
-  while (i < n_atoms)
+  /* This is a bad hack, we always add the
+   * state based on window type in to the state,
+   * even if no state update is pending (since the
+   * state update just means the _NET_WM_STATE prop
+   * changed
+   */
+  
+  if (reread_net_wm_state)
     {
-      if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"))
-        window->priv->is_maximized_vert = TRUE;
-      else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"))
-        window->priv->is_maximized_horz = TRUE;
-      else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_STICKY"))
-        window->priv->is_sticky = TRUE;
-      else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_SHADED"))
-        window->priv->is_shaded = TRUE;
-      else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_SKIP_TASKBAR"))
-        window->priv->skip_taskbar = TRUE;
-      else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_SKIP_PAGER"))
-        window->priv->skip_pager = TRUE;
+      window->priv->is_maximized_horz = FALSE;
+      window->priv->is_maximized_vert = FALSE;
+      window->priv->is_sticky = FALSE;
+      window->priv->is_shaded = FALSE;
+      window->priv->skip_taskbar = FALSE;
+      window->priv->skip_pager = FALSE;
+      window->priv->net_wm_state_hidden = FALSE;
+      
+      atoms = NULL;
+      n_atoms = 0;
+      _wnck_get_atom_list (window->priv->xwindow,
+                           _wnck_atom_get ("_NET_WM_STATE"),
+                           &atoms, &n_atoms);
 
-      ++i;
+      i = 0;
+      while (i < n_atoms)
+        {
+          if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"))
+            window->priv->is_maximized_vert = TRUE;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"))
+            window->priv->is_maximized_horz = TRUE;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_HIDDEN"))
+            window->priv->net_wm_state_hidden = TRUE;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_STICKY"))
+            window->priv->is_sticky = TRUE;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_SHADED"))
+            window->priv->is_shaded = TRUE;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_SKIP_TASKBAR"))
+            window->priv->skip_taskbar = TRUE;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_SKIP_PAGER"))
+            window->priv->skip_pager = TRUE;
+
+          ++i;
+        }
+
+      g_free (atoms);
     }
 
-  g_free (atoms);
+  switch (window->priv->wintype)
+    {
+    case WNCK_WINDOW_DESKTOP:
+    case WNCK_WINDOW_DOCK:
+    case WNCK_WINDOW_TOOLBAR:
+    case WNCK_WINDOW_MENU:
+    case WNCK_WINDOW_UTILITY:
+    case WNCK_WINDOW_SPLASHSCREEN:
+      window->priv->skip_taskbar = TRUE;
+      break;
+      
+    case WNCK_WINDOW_NORMAL:
+    case WNCK_WINDOW_DIALOG:
+    case WNCK_WINDOW_MODAL_DIALOG:
+      break;
+    }
+
+  switch (window->priv->wintype)
+    {
+    case WNCK_WINDOW_DESKTOP:
+    case WNCK_WINDOW_DOCK:
+    case WNCK_WINDOW_TOOLBAR:
+    case WNCK_WINDOW_MENU:
+    case WNCK_WINDOW_SPLASHSCREEN:
+      window->priv->skip_pager = TRUE;
+      break;
+      
+    case WNCK_WINDOW_NORMAL:
+    case WNCK_WINDOW_DIALOG:
+    case WNCK_WINDOW_MODAL_DIALOG:
+    case WNCK_WINDOW_UTILITY:
+      break;
+    }
+
+  /* FIXME we need to recompute this if the window manager changes */
+  if (wnck_screen_net_wm_supports (window->priv->screen,
+                                   "_NET_WM_STATE_HIDDEN"))
+    window->priv->is_minimized = window->priv->net_wm_state_hidden;
+  else
+    window->priv->is_minimized = window->priv->wm_state_iconic;
 }
 
 static void
@@ -1320,6 +1431,101 @@ update_actions (WnckWindow *window)
     WNCK_WINDOW_ACTION_UNMINIMIZE;
 }
 
+static void
+update_wintype (WnckWindow *window)
+{
+  Atom *atoms;
+  int n_atoms;
+  WnckWindowType type;
+  gboolean found_type;
+  
+  if (!window->priv->need_update_wintype)
+    return;
+
+  window->priv->need_update_wintype = FALSE;
+
+  found_type = FALSE;
+  type = WNCK_WINDOW_NORMAL;
+  
+  if (_wnck_get_atom_list (window->priv->xwindow,
+                           _wnck_atom_get ("_NET_WM_WINDOW_TYPE"),
+                           &atoms,
+                           &n_atoms))
+    {
+      int i;
+      
+      i = 0;
+      while (i < n_atoms && !found_type)
+        {
+          /* We break as soon as we find one we recognize,
+           * supposed to prefer those near the front of the list
+           */
+          found_type = TRUE;
+          if (atoms[i] == _wnck_atom_get ("_NET_WM_WINDOW_TYPE_DESKTOP"))
+            type = WNCK_WINDOW_DESKTOP;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_WINDOW_TYPE_DOCK"))
+            type = WNCK_WINDOW_DOCK;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_WINDOW_TYPE_TOOLBAR"))
+            type = WNCK_WINDOW_TOOLBAR;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_WINDOW_TYPE_MENU"))
+            type = WNCK_WINDOW_MENU;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_WINDOW_TYPE_DIALOG"))
+            type = WNCK_WINDOW_DIALOG;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_WINDOW_TYPE_NORMAL"))
+            type = WNCK_WINDOW_NORMAL;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_WINDOW_TYPE_MODAL_DIALOG"))
+            type = WNCK_WINDOW_MODAL_DIALOG;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_WINDOW_TYPE_UTILITY"))
+            type = WNCK_WINDOW_UTILITY;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_WINDOW_TYPE_SPLASHSCREEN"))
+            type = WNCK_WINDOW_SPLASHSCREEN;
+          else
+            found_type = FALSE;
+          
+          ++i;
+        }
+
+      g_free (atoms);
+    }
+
+  if (!found_type)
+    {
+      if (window->priv->transient_for != None)
+        {
+          type = WNCK_WINDOW_DIALOG;
+        }
+      else
+        {
+          type = WNCK_WINDOW_NORMAL;
+        }
+      found_type = TRUE;
+    }
+
+  window->priv->wintype = type;
+}
+
+static void
+update_transient_for (WnckWindow *window)
+{
+  Window parent;
+
+  if (!window->priv->need_update_transient_for)
+    return;
+
+  window->priv->need_update_transient_for = FALSE;
+  
+  if (_wnck_get_window (window->priv->xwindow,
+                        _wnck_atom_get ("WM_TRANSIENT_FOR"),
+                        &parent))
+    {
+      window->priv->transient_for = parent;
+    }
+  else
+    {
+      window->priv->transient_for = None;
+    }
+}
+
 static void
 force_update_now (WnckWindow *window)
 {
@@ -1365,9 +1571,13 @@ force_update_now (WnckWindow *window)
 
   old_state = COMPRESS_STATE (window);
   old_actions = window->priv->actions;
-  
-  update_state (window);
-  update_minimized (window);
+
+  update_transient_for (window); /* wintype needs this to be first */
+  update_wintype (window);
+  update_wm_state (window);
+  update_state (window);     /* must come after the above, since they affect
+                              * our calculated state
+                              */
   update_workspace (window); /* emits signals */
   update_actions (window);
 
